/*
 * This file is part of ElSoldadoYLaElfa.
 *
 * 
 * Miguel Garde Vallés (C) 2012
 * <miguelgardev@gmail.com>
 *
 * 
 * ElSoldadoYLaElfa examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * ElSoldadoYLaElfa examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ElSoldadoYLaElfa examples.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <vector>
#include <list>
#include <map>
#include <set>

#include "IEventManager.h"

//! Clase que gestiona los eventos del juego a aprtir de la intefaz proporcionada por IEventManager.

/**
 *  @author Miguel Garde Vallés
 *  @date 27-03-2012
 *
 *  Registra unos listeners asociandolos a unos eventos, 
 *  de modo que cuando un evento se dispara, se llama al método handleEvent de todos los listeners registrados a él.
 * 
 *  El registro de listeners lo realiza a través de unas tablas hash,
 *  el índice de las tablas es el identificador del evento (el nombre)
 *  y el contenido de las tablas son unas listas de listeners.
 * 
 *  NOTA: Esta clase debería implementarse en el juego, no en el motor,
 *  pero varía tan poco de un juego a otro que por simplicidad se ha decidido meterla aquí.
 * 
 *  Ejemplo:
 *
 *  \code
 *  //inicializamos el manager
 *  IEventManager *evtMgr = IEventManager::Get();
 * 
 *  //registramos los listeners y eventos al gusto
 *  EventListenerPtr listener1 ( new MiListener() );
 *  EventListenerPtr listener2 ( new MiOtroListener() );
 *  evtMgr->addListener( listener1, EventType(MiEvento::nombre) );
 *  evtMgr->addListener( listener1, EventType(MiOtroEvento::nombre) );
 *  evtMgr->addListener( listener2, EventType(MiEvento::nombre) );
 * 
 *  //cuando lo creamos oportuno, disparamos un evento
 *  evtMgr->queueEvent( EventPtr( new MiEvento()) );
 *  //ahora el evento lo recibirán tanto listener1 como listener2.
 *  
 *  \endcode
 */

namespace GardeGames {

	// EventManager Description
	//
	// One implementation of the listener registry
	//



	typedef std::vector<EventListenerPtr> EventListenerList;

	typedef std::vector<EventType>      EventTypeList;

	class EventManager : public IEventManager
	{
	public:

		static EventManager *GetSingletonPtr();

		explicit EventManager( char const * const pName,
										   bool setAsGlobal );

		virtual ~EventManager();

		// --- primary use functions --- make it work ---
		
		// Register a listener for a specific event type, implicitly
		// the event type will be added to the known event types if
		// not already known.
		//
		// The function will return false on failure for any
		// reason. The only really anticipated failure reason is if
		// the input event type is bad ( e.g.: known-ident number
		// with different signature text, or signature text is empty
		// )
		
		bool addListener ( EventListenerPtr const & inListener,
						   EventType const & inType );

		// Remove a listener/type pairing from the internal tables
		//
		// Returns false if the pairing was not found.
		
		bool delListener ( EventListenerPtr const & inListener,
						   EventType const & inType );

		// Fire off event - synchronous - do it NOW kind of thing -
		// analogous to Win32 SendMessage() API.
		//
		// returns true if the event was consumed, false if not. Note
		// that it is acceptable for all event listeners to act on an
		// event and not consume it, this return signature exists to
		// allow complete propogation of that shred of information
		// from the internals of this system to outside uesrs.

		bool trigger ( Event & inEvent ) const;

		// Fire off event - asynchronous - do it WHEN the event
		// system tick() method is called, normally at a judicious
		// time during game-loop processing.
		//
		// returns true if the message was added to the processing
		// queue, false otherwise.

		bool queueEvent ( EventPtr const & inEvent );

		// Find the next-available instance of the named event type
		// and remove it from the processing queue.
		//
		// This may be done up to the point that it is actively being
		// processed ...  e.g.: is safe to happen during event
		// processing itself.
		//
		// if 'allOfType' is input true, then all events of that type
		// are cleared from the input queue.
		//
		// returns true if the event was found and removed, false
		// otherwise

		bool abortEvent ( EventType const & inType,
						  bool allOfType );

		// Allow for processing of any queued messages, optionally
		// specify a processing time limit so that the event
		// processing does not take too long. Note the danger of
		// using this artificial limiter is that all messages may not
		// in fact get processed.
		//
		// returns true if all messages ready for processing were
		// completed, false otherwise (e.g. timeout )

		bool tick ( unsigned long maxMillis );

		// --- information lookup functions ---
		
		// Validate an event type, this does NOT add it to the
		// internal registry, only verifies that it is legal (
		// e.g. either the ident number is not yet assigned, or it is
		// assigned to matching signature text, and the signature
		// text is not empty ).
		
		bool validateType( EventType const & inType ) const {return true;};
		

		// Get the list of listeners associated with a specific event
		// type

		EventListenerList getListenerList ( EventType const & eventType ) const;

		// Get the list of known event types

		EventTypeList getTypeList ( void ) const;

		static EventManager * mEventManager;

	private:

		// one global instance
		typedef std::set< EventType >						EventTypeSet;

		// insert result into event type set
		typedef std::pair< EventTypeSet::iterator, bool >		EventTypeSetIRes;

		// one list per event type ( stored in the map )
		typedef std::list< EventListenerPtr >					EventListenerTable;

		// mapping of event ident to listener list
		typedef std::map< unsigned int, EventListenerTable >	EventListenerMap;

		// entry in the event listener map
		typedef std::pair< unsigned int, EventListenerTable >	EventListenerMapEnt;

		// insert result into listener map
		typedef std::pair< EventListenerMap::iterator, bool >	EventListenerMapIRes;

		// queue of pending- or processing-events
		typedef std::list< EventPtr >							EventQueue;

		enum eConstants
		{
			kNumQueues = 2
		};
		
		EventTypeSet     m_typeList;           // list of registered
												// event types

		EventListenerMap m_registry;           // mapping of event types
												// to listeners

		EventQueue       m_queues[kNumQueues]; // event processing queue,
												// double buffered to
												// prevent infinite cycles

		int               m_activeQueue;        // valid denoting which
												// queue is actively
												// processing, en-queing
												// events goes to the
												// opposing queue

	};
}
