/*
 * This file is part of ElSoldadoYLaElfa.
 *
 * 
 * Miguel Garde Vallés (C) 2012
 * <miguelgardev@gmail.com>
 *
 * 
 * ElSoldadoYLaElfa examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * ElSoldadoYLaElfa examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ElSoldadoYLaElfa examples.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <strstream>


//! Clase auxiliar que crea un hash a partir de un string.

/**
 *  @author Miguel Garde Vallés
 *  @date 27-03-2012
 *
 *  Convierte el nombre de los eventos, un string, a un número entero.
 *  Con esto se consigue que se procese más rápido todo.
 * 
 *  Ejemplo:
 *
 *  \code
 *  //inicializamos el manager
 *  IEventManager *evtMgr = IEventManager::Get();
 * 
 *  //registramos los listeners y eventos al gusto
 *  EventListenerPtr listener1 ( new MiListener() );
 *  evtMgr->addListener( listener1, EventType(MiEvento::nombre) );
 *  
 *  \endcode
 */

namespace GardeGames {

	char const * const kpWildcardEventType = "*";

	class EventType
	{
	public:

		explicit EventType( char const * const pIdentStr )
			: m_ident( hash_name(pIdentStr) ),
			m_identStr(pIdentStr)
		{}

		unsigned long getIdent() const
		{
			return reinterpret_cast<unsigned long>(m_ident);
		}

		char const * const getStr() const
		{
			return m_identStr;
		}

		bool operator< ( EventType const & o ) const
		{
			bool r = ( getIdent() < o.getIdent() );
			return r;
		}

		bool operator== ( EventType const & o ) const
		{
			bool r = ( getIdent() == o.getIdent() );
			return r;
		}

		void* EventType::hash_name( char const * pIdentStr )
	{
		// Relatively simple hash of arbitrary text string into a
		// 32-bit identifier Output value is
		// input-valid-deterministic, but no guarantees are made
		// about the uniqueness of the output per-input
		//
		// Input value is treated as lower-case to cut down on false
		// separations cause by human mistypes. Sure, it could be
		// construed as a programming error to mix up your cases, and
		// it cuts down on permutations, but in Real World Usage
		// making this text case-sensitive will likely just lead to
		// Pain and Suffering.
		//
		// This code lossely based upon the adler32 checksum by Mark
		// Adler and published as part of the zlib compression
		// library sources.

		// largest prime smaller than 65536
		unsigned long BASE = 65521L;

		// NMAX is the largest n such that 255n(n+1)/2 +
		// (n+1)(BASE-1) <= 2^32-1
		unsigned long NMAX = 5552;

	#define DO1(buf,i)  {s1 += tolower(buf[i]); s2 += s1;}
	#define DO2(buf,i)  DO1(buf,i); DO1(buf,i+1);
	#define DO4(buf,i)  DO2(buf,i); DO2(buf,i+2);
	#define DO8(buf,i)  DO4(buf,i); DO4(buf,i+4);
	#define DO16(buf)   DO8(buf,0); DO8(buf,8);

		if (pIdentStr == NULL)
			return NULL;

		if ( strcmp( pIdentStr, kpWildcardEventType ) == 0 )
			return 0;
		
		unsigned long s1 = 0;
		unsigned long s2 = 0;

		for ( size_t len = strlen( pIdentStr ); len > 0 ; )
		{
			unsigned long k = len < NMAX ? len : NMAX;

			len -= k;

			while (k >= 16)
			{
				DO16(pIdentStr);
				pIdentStr += 16;
				k -= 16;
			}
			
			if (k != 0) do
			{
				s1 += *pIdentStr++;
				s2 += s1;
			} while (--k);
			
			s1 %= BASE;
			s2 %= BASE;
		}

	#pragma warning(push)
	#pragma warning(disable : 4312)

		return reinterpret_cast<void *>( (s2 << 16) | s1 );

	#pragma warning(pop)
	#undef DO1
	#undef DO2
	#undef DO4
	#undef DO8
	#undef DO16
	}

	private:

		// note: m_ident is stored as a void* not an int, so that in
		// the debugger it will show up as hex-values instead of
		// integer values. This is a bit more representative of what
		// we're doing here and makes it easy to allow external code
		// to assign event types as desired.

		void *       m_ident;
		char const * m_identStr;
		
	};
}
