/*
 * This file is part of ElSoldadoYLaElfa.
 *
 * 
 * Miguel Garde Vallés (C) 2012
 * <miguelgardev@gmail.com>
 *
 * 
 * ElSoldadoYLaElfa examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * ElSoldadoYLaElfa examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ElSoldadoYLaElfa examples.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <OIS.h>
#include <Ogre.h>
#include "GameAction.h"

//! Clase que gestiona la entrada de teclado y ratón.


/**
 *  @author Miguel Garde Vallés
 *  @date 27-03-2012
 *
 *  Registra las teclas pulsadas y los movimientos de ratón realizados.
 *  También mapea las GameAction con las teclas y el ratón
 * 
 *  La clase tiene dos arrays de GameAction:
 *  en uno, los índices son el keyCode de las teclas, de modo que si se pulsa la 'w'
 *  se le dice a la GameAction indexada en la posición de la 'w' (si la hay) que debe activarse.
 * 
 *  Lo mismo ocurre para el ratón, hay un array cuyos índices son unas constantes 
 *  predefinidas en la propia clase para indicar si se ha hecho click, si se ha movido hacia abajo el ratón, etc.
 *  
 *  La clase implementa keyListener y mouseListener para tener constancia de los cambios.
 * 
 *  Ejemplo:
 *
 *  \code
 *  //el inputManager se inicializa en el GameManager, de modo que no hace falta hacerlo manualmente
 * 
 *  //se asocian a una o varias teclas a las GameAction
 *  inputManager->mapToKey(gameAction1,OIS::KC_W);
 *  inputManager->mapToKey(gameAction1,OIS::KC_UP);
 *  inputManager->mapToMouse(gameAction2,InputManager::MOUSE_BUTTON_LEFT);
 *  
 *  \endcode
 */

namespace GardeGames
{
	class InputManager : public OIS::KeyListener, public OIS::MouseListener//, public OIS::JoyStickListener
    {
		public:
			enum MouseCodes {
				MOUSE_MOVE_LEFT = 0,
				MOUSE_MOVE_RIGHT = 1,
				MOUSE_MOVE_UP = 2,
				MOUSE_MOVE_DOWN = 3,
				MOUSE_WHEEL_UP = 4,
				MOUSE_WHEEL_DOWN = 5,
				MOUSE_BUTTON_LEFT = 6,
				MOUSE_BUTTON_RIGHT = 7,
				MOUSE_BUTTON_MIDDLE = 8,

				NUM_MOUSE_CODES = 9
			};
			
			
			__declspec(dllexport)	InputManager();
			//__declspec(dllexport)	InputManager(Ogre::RenderWindow *window);
			__declspec(dllexport)	~InputManager();
			__declspec(dllexport)	void setKeyboard(OIS::Keyboard *mKeyboard);
			__declspec(dllexport)	void setMouse(OIS::Mouse *mMouse);
			__declspec(dllexport)	void mapToKey(GameAction *accion, OIS::KeyCode key);
			__declspec(dllexport)	void mapToMouse(GameAction *accion, int codigo);
			__declspec(dllexport)	void capture();
			__declspec(dllexport)	void clearAllMaps();

			__declspec(dllexport)	int getMouseX();
			__declspec(dllexport)	int getMouseY();
			__declspec(dllexport)	int getMouseRelativeX();
			__declspec(dllexport)	int getMouseRelativeY();

		private:
			enum Teclas {
				NUM_TECLAS = 600
			};
			GameAction *accionesTeclas[NUM_TECLAS];
			GameAction *accionesRaton[NUM_MOUSE_CODES];
			OIS::KeyListener* keyListener;
			OIS::Keyboard *teclado;
			OIS::Mouse *raton;

			int mRatonX;
			int mRatonY;

			int mDesplazamientoX;
			int mDesplazamientoY;

			void mouseHelper(int codeNeg, int codePos,int amount);
			GameAction* getKeyAction(OIS::KeyCode key);
			int getMouseButtonCode(OIS::MouseButtonID id);
			GameAction* getMouseButtonAction(OIS::MouseButtonID id);
			
			bool keyPressed( const OIS::KeyEvent &e );
			bool keyReleased( const OIS::KeyEvent &e );
			
			bool mouseMoved( const OIS::MouseEvent &arg );
			bool mousePressed( const OIS::MouseEvent &arg, OIS::MouseButtonID id );
			bool mouseReleased( const OIS::MouseEvent &arg, OIS::MouseButtonID id  );

	};
}
