/*
 * This file is part of ElSoldadoYLaElfa.
 *
 * 
 * Miguel Garde Vallés (C) 2012
 * <miguelgardev@gmail.com>
 *
 * 
 * ElSoldadoYLaElfa examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * ElSoldadoYLaElfa examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ElSoldadoYLaElfa examples.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "Escenario.h"
#include "Elemento.h"
#include "Ogre.h"
#include "Terrain/OgreTerrain.h"
#include "Terrain/OgreTerrainGroup.h"

#include "GameActor.h"
#include "Attack.h"

#define MAX_ACTORS_PER_MAP 200

using namespace std;

//! Clase que gestiona la carga y descarga de recursos en el escenario.


/**
 *  @author Miguel Garde Vallés
 *  @date 14-01-2012
 *
 *  Gestiona todos los recursos del escenario en dos secciones:
 *  en primer lugar gestiona todo lo relacionado con el editor de escenarios,
 *  hace de pasarela entre el editor y la clase Escenario.
 *  Así mismo, gestiona todo lo relacionado con el in-game,
 *  desde crear y destruir los nodos, hasta la contrucción de overlays y barras de vida;
 *  pasando por la creación del terreno mediante mapas de alturas.
 *  
 *  También trata todo lo relacionado con el escenario en cuanto al ray casting, la altura de un punto, etc.
 */

class GestorEscenario
{
public:
	GestorEscenario(void);
	~GestorEscenario(void);
	
	enum tipoElemento {
		CASA,
		OGRO,
		NINJA,
		ROBOT,
		NUM_ELEMENTOS
	};

	static GestorEscenario *getSingletonPtr();

	Ogre::SceneManager* getSceneManager();

	//A�ade o quita un objeto que haya sido colocado en el escenario
	Ogre::SceneNode* GestorEscenario::addElement(tipoElemento tipo, Ogre::Vector3 pos);
	void removeElement(Ogre::SceneNode* objeto);

	//establece la ubicacion en la que aparecera el jugador
	void setInitialPosition(Ogre::Vector3 pos);

	//guarda, carga o crea un escenario a partir de un nombre dado
	bool saveEscenario();
	bool loadEscenario(char* nombre);
	bool createNewEscenario(char* archivo, char* nombre);

	bool isLoading();

	//establece el terreno (suelo) y cielo del escenario a partir de los nombres de las imagenes
	bool setTerreno(std::string);
	bool setCielo(std::string);

	//cambia la ambientacion visual
	void setNiebla(Ogre::FogMode tipo,Ogre::ColourValue color,float densidad,float comienzo, float fin);
	void setLuz(Ogre::ColourValue difuso, Ogre::ColourValue specular, Ogre::ColourValue ambient);

	Ogre::Camera* getCamera();
	Ogre::SceneNode* getAttachedCameraNode();
	Ogre::RaySceneQuery* getRay();
	Ogre::SceneQueryResult getVolumeResult(Ogre::Ray topLeft, Ogre::Ray topRight,
											Ogre::Ray botLeft, Ogre::Ray botRight);

	Ogre::TerrainGroup::RayResult getRayCollision(Ogre::Ray rayo);
	float getHeightAtPosition(float x, float y, float z);
	float getHeightAtPosition(float x, float z);
	float getHeightAtPosition(Ogre::Vector3 &v);

//	NxOgre::World* mWorld;
//	NxOgre::Scene* mScene;
//	Critter::RenderSystem* mRenderSystem;
        
        GameActor* getGameActorAt(int index);
	GameActor* createGameActor(GameActor::ActorType type,Ogre::Vector3 position);
        void destroyGameActor(GameActor* gameActor);
        
        Ogre::Billboard* createLifeBar(GameActor* gameActor);
        void destroyLifeBar(GameActor* gameActor);
        
        Attack* createAttack(Attack::ATTACK_TYPE type, GameActor* owner, Ogre::Vector3 direction, float heightAttack); 
        

private:
	int mPosRatonX;
	int mPosRatonY;

	Escenario* mEscenario;
	int mNumElementos;
        
        int mNumActors;
	GameActor *mListGameActor[200];
        
        int mNumAttacks;
        Attack* mListAttacks[200];
	
	Ogre::SceneManager*			mSceneManager;
	
	Ogre::TerrainGlobalOptions* mTerrainGlobals;
    Ogre::TerrainGroup*			mTerrainGroup;
	bool						mTerrainsImported;
	Ogre::Camera*				mCamera;

	static GestorEscenario *gestor;

	void limpiarGestor();

	void createTerrain(Ogre::Light *light, std::string nombreImagen);
	void loadTerrainGeometry(const Ogre::String& name, float* data, Ogre::uint16 size, Ogre::Real worldSize,
									Ogre::Real minHeight, Ogre::Real maxHeight, const Ogre::Vector3& position);
	void initBlendMaps(Ogre::Terrain* terrain);
	void defineTerrain(long x, long y, std::string nombreImagen);
	void configureTerrainDefaults(Ogre::Light* light);
	Ogre::Image getTerrainImage(bool flipX, bool flipY, std::string nombreImagen);
	void setupHeightFields();

	void createScene();
	void destroyScene();
	
	GameActor* createWolf(char* nombre, Ogre::Vector3 position);
	GameActor* createDeer(char* nombre, Ogre::Vector3 position);
        GameActor* createPlayer(char* nombre, Ogre::Vector3 position);
           GameActor* createDragon(char* nombre, Ogre::Vector3 position);
        
        Attack* createMeleeAttack(char* nombre, GameActor* owner, Ogre::Vector3 direction, float heightAttack); 
        Attack* createFireAttack(char* nombre, GameActor* owner, Ogre::Vector3 direction, float heightAttack); 
};
